from mw2fcitx.exporters.opencc import export


def test_opencc_exporter():
    assert (
        export(["测试"]) == "测试\tce'shi\t0\n"
    )

    assert (
        export([
            "测试",
            "琴吹䌷"  # outloudvi/mw2fcitx#16
        ]) == "测试\tce'shi\t0\n"
        "琴吹䌷\tqin'chui'chou\t0\n"
    )

    assert (
        export([
            "测试",
            "无效:词条"
        ]) == "测试\tce'shi\t0\n"

    )



import json
from pypinyin import lazy_pinyin
import opencc
from ..logger import console

DEFAULT_PLACEHOLDER = "_ERROR_"


def manual_fix(text, table):
    if text in table:
        return table[text]
    return None


def export(words, **kwargs):
    result = ""
    converter = opencc.OpenCC('t2s.json')
    fixfile = kwargs.get("fixfile")
    if fixfile is not None:
        table = json.load(open(fixfile, "r", encoding="utf-8"))
    count = 0
    for line in words:
        line = line.rstrip("\n")
        pinyins = lazy_pinyin(line, errors=lambda x: DEFAULT_PLACEHOLDER)
        if DEFAULT_PLACEHOLDER in pinyins:
            # The word is not fully converable
            continue
        pinyin = "'".join(pinyins)
        if pinyin == line:
            # print("Failed to convert, ignoring:", pinyin, file=sys.stderr)
            continue

        if fixfile is not None:
            fixed_pinyin = manual_fix(line, table)
            if fixed_pinyin is not None:
                pinyin = fixed_pinyin
                console.debug(f"Fixing {line} to {pinyin}")

        result += "\t".join((converter.convert(line), pinyin, "0"))
        result += "\n"
        count += 1
        if count % 1000 == 0:
            console.debug(str(count) + " converted")

    if count % 1000 != 0 or count == 0:
        console.debug(str(count) + " converted")
    return result



from mw2fcitx.main import inner_main
import pytest


def test_inner_main_name_without_py():
    inner_main(['-c', 'tests/cli/conf_one'])


def test_inner_main_name_with_py():
    inner_main(['-c', 'tests/cli/conf_one.py'])


def test_single_generator():
    inner_main(['-c', 'tests/cli/conf_single_generator'])


def test_local():
    inner_main(['-c', 'tests/cli/conf_local'])
    with open("test_local_result.dict.yml", "r") as f:
        assert (f.read() == """---
name: e2etest_local
version: "0.1"
sort: by_weight
...
初音未来	chu yin wei lai
""")


def test_err_no_path():
    with pytest.raises(SystemExit):
        inner_main(['-c', 'tests/cli/conf_err_no_path'])



import time
from functools import wraps

# https://www.saltycrane.com/blog/2009/11/trying-out-retry-decorator-python/


def retry(tries=4, delay=3, backoff=2):

    def deco_retry(f):

        @wraps(f)
        def f_retry(*args, **kwargs):
            mtries, mdelay = tries, delay
            while mtries > 1:
                try:
                    return f(*args, **kwargs)
                except Exception as e:
                    print("%s, Retrying in %d seconds..." % (str(e), mdelay))
                    time.sleep(mdelay)
                    mtries -= 1
                    mdelay *= backoff
            return f(*args, **kwargs)

        return f_retry

    return deco_retry



from http.client import HTTPException
import json
import sys
from json.decoder import JSONDecodeError
from os import access, R_OK, fdopen
from tempfile import mkstemp
from urllib.error import HTTPError, URLError
from urllib.request import Request, urlopen
from urllib.parse import quote_plus

from .logger import console
from .retry import retry


class StatusError(Exception):

    def __init__(self, code):
        super().__init__(f"HTTP status is {code}")


@retry()
def open_request(url):
    return urlopen(
        Request(
            url,
            headers={
                "User-Agent":
                    "MW2Fcitx/1.0; github.com/outloudvi/fcitx5-pinyin-moegirl"
            }))


def fetch_as_json(url):
    for _ in range(3):
        try:
            res = open_request(url)
            if res.status == 200:
                return json.loads(res.read())
        except (JSONDecodeError, HTTPError) as e:
            console.error(f"Fetch error {str(e)}, retrying")
            continue
    console.error(f"Error fetching URL {url}")
    raise StatusError(res.status)


def save_to_partial(partial_path, titles, apcontinue):
    ret = {"apcontinue": apcontinue, "titles": titles}
    try:
        with open(partial_path, "w", encoding="utf-8") as fp:
            fp.write(json.dumps(ret, ensure_ascii=False))
        console.debug(f"Partial session saved to {partial_path}")
    except Exception as e:
        console.error(str(e))


def resume_from_partial(partial_path):
    if not access(partial_path, R_OK):
        console.warn(f"Cannot read partial session: {partial_path}")
        return [[], None]
    try:
        partial_data = json.load(open(partial_path, "r", encoding="utf-8"))
        titles = partial_data.get("titles", [])
        apcontinue = partial_data.get("apcontinue", None)
        return [titles, apcontinue]
    except Exception as e:
        console.error(str(e))
        console.error("Failed to parse partial session")
        return [[], None]


def fetch_all_titles(api_url, **kwargs):
    limit = kwargs.get("api_title_limit") or kwargs.get("title_limit") or -1
    console.debug(f"Fetching titles from {api_url}" +
                  (f" with a limit of {limit}" if limit != -1 else ""))
    titles = []
    partial_path = kwargs.get("partial")
    fetch_url = api_url + "?action=query&list=allpages&format=json"
    if partial_path is not None:
        console.info(f"Partial session will be saved/read: {partial_path}")
        [titles, apcontinue] = resume_from_partial(partial_path)
        if apcontinue is not None:
            fetch_url = api_url + \
                f"?action=query&list=allpages&format=json&aplimit=max&apcontinue={quote_plus(apcontinue)}"
            console.info(
                f"{len(titles)} titles found. Continuing from {apcontinue}")
    try:
        data = fetch_as_json(fetch_url)
    except (StatusError, HTTPError):
        # It somehow cannot proceed
        if len(titles) == 0:
            raise Exception("Failed to retrieve any title")
        else:
            # Failed during fetching
            [fd, fpath] = mkstemp(prefix="mwf-titles")
            with fdopen(fd, 'w') as f:
                f.write('\n'.join(titles))
            console.error(
                f"Fetch halted due to exceptions. Check '{fpath}' for current progress."
            )
            raise Exception(
                f"Failed during fetching the titles: {len(titles)} fetched and failed at '{apcontinue}'."
            )
    breakNow = False
    while True:
        for i in map(lambda x: x["title"], data["query"]["allpages"]):
            titles.append(i)
            if limit != -1 and len(titles) >= limit:
                breakNow = True
                break
        console.debug(f"Got {len(titles)} pages")
        if breakNow:
            break
        if "continue" in data:
            try:
                apcontinue = data["continue"]["apcontinue"]
                data = fetch_as_json(
                    api_url +
                    f"?action=query&list=allpages&format=json&aplimit=max&apcontinue={quote_plus(apcontinue)}"
                )
            except (HTTPException, TimeoutError, URLError) as e:
                console.error(str(e))
                if partial_path:
                    save_to_partial(partial_path, titles, apcontinue)
                sys.exit(1)
            except KeyboardInterrupt:
                console.error("Keyboard interrupt received. Stopping.")
                if partial_path:
                    save_to_partial(partial_path, titles, apcontinue)
                sys.exit(1)
        else:
            break
    console.info("Finished.")
    return titles



from copy import deepcopy
from typing import List


def normalize(word):
    return word.strip()


def sanitize(obj):
    res = deepcopy(obj)
    typ = type(res)
    if typ == type(sanitize):  # function
        func_name = res.__name__ or "lambda"
        return f"[func {func_name}]"
    if typ == type({}):  # object
        for i in res.keys():
            res[i] = sanitize(res[i])
    elif typ == type([]):  # list
        fin = []
        for i in res:
            fin.append(sanitize(i))
        res = fin
    elif typ == type(1) or typ == type("1"):  # number
        return str(res)
    else:  # whatever
        return "[" + str(type(res)) + "]"
    return res


def smart_rewrite(config_object):

    # If `generator` is not a list, make it a list
    generators = config_object["generator"]
    if not isinstance(generators, list):
        config_object["generator"] = [generators]

    # If `title_file_path` is not a list, make it a list
    title_file_path = config_object["source"].get("file_path")
    if isinstance(title_file_path, str):
        config_object["source"]['file_path'] = [title_file_path]

    return config_object


def is_libime_used(config):
    generators = config.get('generator') or []
    for i in generators:
        if i.get("use") == "pinyin":
            return True
    return False


def dedup(arr: List[str]):
    return list(set(arr))



# pylint: disable=import-outside-toplevel

import os
import sys

from .fetch import fetch_all_titles
from .utils import dedup
from .logger import console


class MWFPipeline():

    def __init__(self, api_path=""):
        self.api_path = api_path
        self.titles = []
        self.words = []
        self.exports = ""
        self.dict = ""

    def load_titles(self, titles, limit=-1, replace=False):
        if isinstance(titles, str):
            titles = titles.split("\n")
        if limit >= 0:
            titles = titles[:limit]
        if replace:
            self.titles = titles
        else:
            self.titles.extend(titles)
        console.debug(f"{len(titles)} title(s) imported.")
        self.words = self.titles

    def write_titles_to_file(self, filename):
        try:
            with open(filename, "w") as file:
                file.write("\n".join(self.titles))
        except Exception as e:
            console.error(f"File {filename} is not writable: {str(e)}")
            sys.exit(1)

    def post_load(self, **kwargs):
        if kwargs.get("output"):
            self.write_titles_to_file(kwargs.get("output"))

    def load_titles_from_file(self, filename, **kwargs):
        limit = kwargs.get("file_title_limit") or kwargs.get(
            "title_limit") or -1
        if not os.access(filename, os.R_OK):
            console.error(
                f"File {filename} is not readable; remove this parameter (\"file_path\") or provide a readable file"
            )
            sys.exit(1)
        self.load_titles(open(filename, "r").read(), limit=limit)

    def fetch_titles(self, **kwargs):
        titles = fetch_all_titles(self.api_path, **kwargs)
        self.load_titles(titles)
        self.post_load(**kwargs)

    def reset_words(self):
        self.words = self.titles

    def convert_to_words(self, pipelines):
        console.debug(f"Running {len(pipelines)} pipelines")
        cnt = 0
        if callable(pipelines):
            pipelines = [pipelines]
        titles = self.titles
        for i in pipelines:
            cnt += 1
            console.debug(
                f"Running pipeline {cnt}/{len(pipelines)} ({i.__name__ or 'anonymous function'}')"
            )
            titles = i(titles)
        console.debug(f"Deduplicating {len(titles)} items")
        self.words = dedup(titles)
        console.debug(
            f"Deduplication completed. {len(self.words)} items left.")

    def export_words(self, converter="opencc", **kwargs):
        if converter == "opencc":
            console.debug(f"Exporting {len(self.words)} words with OpenCC")
            from mw2fcitx.exporters.opencc import export
            self.exports = export(self.words, **kwargs)
        elif callable(converter):
            console.debug(
                f"Exporting {len(self.words)} words with custom converter")
            self.exports = converter(self.words, **kwargs)
        else:
            console.error(f"No such exporter: {converter}")

    def generate_dict(self, generator="pinyin", **kwargs):
        if generator == "pinyin":
            from mw2fcitx.dictgen import pinyin
            dest = kwargs.get("output")
            if not dest:
                console.error(
                    "Dictgen 'pinyin' can only output to files.\n" +
                    "Please give the file path in the 'output' argument.")
                return
            pinyin(self.exports, **kwargs)
        elif generator == "rime":
            from mw2fcitx.dictgen import rime
            self.dict = rime(self.exports, **kwargs)
        else:
            console.error(f"No such dictgen: {generator}")



import json
import os
import re
import shutil
import sys
from argparse import ArgumentParser
from importlib import import_module

from .build_dict import build
from .const import LIBIME_BIN_NAME, LIBIME_REPOLOGY_URL
from .logger import console
from .utils import sanitize, is_libime_used, smart_rewrite


def get_args(args):
    parser = ArgumentParser(
        usage="Fetch titles from online and generate a dictionary.")
    parser.add_argument("-c",
                        "--config",
                        dest="config",
                        default="config.py",
                        help="configuration file location")
    parser.add_argument("-n",
                        "--name",
                        dest="name",
                        default="exports",
                        help="configuration object name")

    return parser.parse_args(args)


def try_file(file):
    console.debug(f"Finding config file: {file}")
    if not os.access(file, os.R_OK):
        console.error("File ({}) not readable.")
        return False
    file_realpath = os.path.realpath(file)
    console.debug(f"Config file path: {file_realpath}")
    file_path = os.path.dirname(file_realpath)
    file_name = os.path.basename(file_realpath)
    module_name = re.sub(r"\.py$", "", file_name)
    config_file = False
    try:
        sys.path.insert(1, file_path)
        config_file = import_module(module_name)
    except Exception as e:
        console.error(f"Error reading config: {str(e)}")
        return False
    finally:
        sys.path.remove(file_path)
    return config_file


def inner_main(args):
    options = get_args(args)
    file = options.config
    objname = options.name
    if file.endswith(".py"):
        config_base = try_file(file)
        if not config_base:
            # I don't think it works... but let's put it here
            config_base = try_file(file + ".py")
    else:
        config_base = try_file(file + ".py")
    if not config_base:
        filename = f"{file}, {file}.py" if file.endswith("py") else file
        console.error(f"Config file {filename} not found or not readable")
        sys.exit(1)
    console.debug(f"Parsing config file: {file}")
    if objname not in dir(config_base):
        console.error(
            f"Exports not found. Please make sure your config in in a object called '{objname}'."
        )
        sys.exit(1)
    config_object = getattr(config_base, objname)
    console.debug("Config load:")
    displayable_config_object = sanitize(config_object)
    if not isinstance(config_object, object):
        console.error("Invalid config")
        sys.exit(1)
    console.debug(
        json.dumps(displayable_config_object, indent=2, sort_keys=True))
    config_object = smart_rewrite(config_object)
    if is_libime_used(config_object) and shutil.which(LIBIME_BIN_NAME) is None:
        console.warn(
            f"You are trying to generate fcitx dictionary, while {LIBIME_BIN_NAME} doesn't seem to exist. This might cause issues. Please install libime: {LIBIME_REPOLOGY_URL}"
        )
    build(config_object)


def main():
    inner_main(sys.argv[1:])
